@@ -18,7 +18,7 @@ use File::Basename;
 
 # globale persistent variables for speedy
 use vars qw($cfg $probes $VERSION $havegetaddrinfo $cgimode);
-$VERSION="1.30";
+$VERSION="1.34";
 
 # we want opts everywhere
 my %opt;
@@ -119,113 +119,161 @@ sub sendmail ($$$){
 	$smtp->datasend($body);
 	$smtp->dataend();
 	$smtp->quit;
-    } elsif ($cfg->{General}{sendmail} or -f "/usr/lib/sendmail"){
+    } elsif ($cfg->{General}{sendmail} or -x "/usr/lib/sendmail"){
 	open (M, "|-") || exec (($cfg->{General}{sendmail} || "/usr/lib/sendmail"),"-f",$from,$to);
 	print M $body;
 	close M;
     }
 }
 
+sub sendsnpp ($$){
+   my $to = shift;
+   my $msg = shift;
+   if ($cfg->{General}{snpphost}){
+        my $snpp = Net::SNPP->new($cfg->{General}{snpphost}, Timeout => 60);
+        $snpp->send( Pager => $to,
+                     Message => $msg) || $opt{debug} and warn "ERROR - ". $snpp->message;
+        $snpp->quit;
+    }
+}
+
 sub init_alerts ($){
     my $cfg = shift;
     foreach my $al (keys %{$cfg->{Alerts}}) {
 	my $x = $cfg->{Alerts}{$al};
         next unless ref $x eq 'HASH';
-	my $sub_front = <<SUB;
+	if ($x->{type} eq 'matcher'){
+	    $x->{pattern} =~ /(\S+)\((.+)\)/
+		or die "ERROR Alert $al pattern entry '$_' is invalid\n";
+	    my $matcher = $1;
+	    my $arg = $2;
+	    eval 'require matchers::'.$matcher;
+	    die "Matcher '$matcher' could not be loaded: $@\n" if $@;
+	    my $hand;
+	    eval "\$hand = matchers::$matcher->new($arg)";
+  	    die "ERROR: Matcher '$matcher' could not be instantiated\nwith arguments $arg:\n$@\n" if $@;
+	    $x->{length} = $hand->Length;
+	    $x->{sub} = sub { $hand->Test(shift) } ;
+	} else {
+	    my $sub_front = <<SUB;
 sub { 
     my \$d = shift;
     my \$y = \$d->{$x->{type}};
     for(1){
 SUB
-        my $sub;
-        my $sub_back = "        return 1;\n    }\n    return 0;\n}\n";
-	my @ops = split /\s*,\s*/, $x->{pattern};
-	$x->{length} = scalar grep /^[!=><]/, @ops;
-        my $multis = scalar grep /^[*]/, @ops;
-        my $it = "";
-        for(1..$multis){
-           my $ind = "    " x ($_-1);
-           $sub .= <<FOR;
+	    my $sub;
+	    my $sub_back = "        return 1;\n    }\n    return 0;\n}\n";
+	    my @ops = split /\s*,\s*/, $x->{pattern};
+	    $x->{length} = scalar grep /^[!=><]/, @ops;
+	    my $multis = scalar grep /^[*]/, @ops;
+	    my $it = "";
+	    for(1..$multis){
+		my $ind = "    " x ($_-1);
+		$sub .= <<FOR;
 $ind        my \$i$_;
 $ind        for(\$i$_=0; \$i$_<\$imax$_;\$i$_++){
 FOR
-        };
-        my $i = - $x->{length};
-        my $incr = 0;
-	for (@ops) {
-            my $extra = "";
-            $it = "    " x $multis;
-	    for(1..$multis){
-              $extra .= "-\$i$_";
-            };
-	    /^(==|!=|<|>|<=|>=|\*)(\d+(?:\.\d*)?|U|S|\d*\*)(%?)$/
-		or die "ERROR Alert $al pattern entry '$_' is invalid\n";
-	    my $op = $1;
-	    my $value = $2;
-            my $perc = $3;
-            if ($op eq '*') {
-		if ($value =~ /^([1-9]\d*)\*$/) {
-		    $value = $1;
-		    $x->{length} += $value;
-                    $sub_front .= "        my \$imax$multis = $value;\n";
-                    $sub_back .=  "\n";
-                    $sub .= <<FOR;
+	    };
+	    my $i = - $x->{length};
+	    my $incr = 0;
+	    for (@ops) {
+		my $extra = "";
+		$it = "    " x $multis;
+		for(1..$multis){
+		    $extra .= "-\$i$_";
+		};
+		/^(==|!=|<|>|<=|>=|\*)(\d+(?:\.\d*)?|U|S|\d*\*)(%?)$/
+		    or die "ERROR Alert $al pattern entry '$_' is invalid\n";
+		my $op = $1;
+		my $value = $2;
+		my $perc = $3;
+		if ($op eq '*') {
+		    if ($value =~ /^([1-9]\d*)\*$/) {
+			$value = $1;
+			$x->{length} += $value;
+			$sub_front .= "        my \$imax$multis = $value;\n";
+			$sub_back .=  "\n";
+			$sub .= <<FOR;
 $it        last;
 $it    }
 $it    return 0 if \$i$multis >= \$imax$multis;
 FOR
-                
-                    $multis--;
+			
+			$multis--;
                     next;
-		} else {
-		    die "ERROR: multi-match operator * must be followed by Number* in Alert $al definition\n";
-		}
-	    } elsif ($value eq 'U') {
-		if ($op eq '==') {
-		    $sub .= "$it        next if defined \$y->[$i$extra];\n";
+		    } else {
+			die "ERROR: multi-match operator * must be followed by Number* in Alert $al definition\n";
+		    }
+		} elsif ($value eq 'U') {
+		    if ($op eq '==') {
+			$sub .= "$it        next if defined \$y->[$i$extra];\n";
 		} elsif ($op eq '!=') {
 		    $sub .= "$it        next unless defined \$y->[$i$extra];\n";
 		} else {
 		    die "ERROR: invalid operator $op in connection U in Alert $al definition\n";
 		}
-	    } elsif ($value eq 'S') {
-		if ($op eq '==') {
-		    $sub .= "$it        next unless defined \$y->[$i$extra] and \$y->[$i$extra] eq 'S';\n";
-		} else {
-		    die "ERROR: S is only valid with == operator in Alert $al definition\n";
+		} elsif ($value eq 'S') {
+		    if ($op eq '==') {
+			$sub .= "$it        next unless defined \$y->[$i$extra] and \$y->[$i$extra] eq 'S';\n";
+		    } else {
+			die "ERROR: S is only valid with == operator in Alert $al definition\n";
 		}
-	    } elsif ($value eq '*') {
-		if ($op ne '==') {
-		    die "ERROR: operator $op makes no sense with * in Alert $al definition\n";
-		} # do nothing else ...
-	    } else {
-		if ( $x->{type} eq 'loss') {
-                    die "ERROR: loss should be specified in % (alert $al pattern)\n" unless $perc eq "%";
-		    $value = $value / 100 * $cfg->{Database}{pings};
+		} elsif ($value eq '*') {
+		    if ($op ne '==') {
+			die "ERROR: operator $op makes no sense with * in Alert $al definition\n";
+		    } # do nothing else ...
+		} else {
+		    if ( $x->{type} eq 'loss') {
+			die "ERROR: loss should be specified in % (alert $al pattern)\n" unless $perc eq "%";
+			$value = $value / 100 * $cfg->{Database}{pings};
 		} elsif ( $x->{type} eq 'rtt' ) {
 		    $value /= 1000;
 		} else {
 		    die "ERROR: unknown alert type $x->{type}\n";
 		}
-		$sub .= <<IF;
+		    $sub .= <<IF;
 $it        next unless defined \$y->[$i$extra]
 $it                        and \$y->[$i$extra] =~ /^\\d/
 $it                        and \$y->[$i$extra] $op $value;
 IF
+		}
+		$i++;
 	    }
-	    $i++;
-	}
-        $sub_front .= "$it        next if scalar \@\$y < $x->{length} ;\n";
-	warn <<COMP if $opt{debug};
+	    $sub_front .= "$it        next if scalar \@\$y < $x->{length} ;\n";
+	    warn <<COMP if $opt{debug};
 ### Compiling alert detector pattern '$al'
 ### $x->{pattern}
 $sub_front$sub$sub_back
 COMP
-	$x->{sub} = eval ( $sub_front.$sub.$sub_back );
-	die "ERROR: compiling alert pattern $al ($x->{pattern}): $@\n" if $@;
+	    $x->{sub} = eval ( $sub_front.$sub.$sub_back );
+	    die "ERROR: compiling alert pattern $al ($x->{pattern}): $@\n" if $@;
+	}
     }
 }
 
+
+sub check_filter ($) {
+    my $name = shift;
+    # if there is a filter do neither schedule these nor make rrds
+    if ($opt{filter} && scalar @{$opt{filter}}){
+         my $ok = 0;
+         for (@{$opt{filter}}){
+            /^\!(.+)$/ && do {
+    	        my $rx = $1;
+                $name !~ /^$rx/ && do{ $ok = 1};
+                next;
+            };
+            /^(.+)$/ && do {
+	        my $rx = $1;
+                $name =~ /^$rx/ && do {$ok = 1};
+                next;
+            }; 
+         }  
+         return $ok;
+      };
+      return 1;
+}
+
 sub init_target_tree ($$$$$$$$); # predeclare recursive subs
 sub init_target_tree ($$$$$$$$) {
     my $cfg = shift;
@@ -259,7 +307,7 @@ sub init_target_tree ($$$$$$$$) {
 	$tree->{alerts} = [ split(/\s*,\s*/, $tree->{alerts}) ] unless ref $tree->{alerts} eq 'ARRAY';
 	$tree->{fetchlength} = 0;
  	foreach my $al (@{$tree->{alerts}}) {
-	    die "ERROR: alert $al ($name) is not defined"
+	    die "ERROR: alert $al ($name) is not defined\n"
 		unless defined $cfg->{Alerts}{$al};
 	    $tree->{fetchlength} = $cfg->{Alerts}{$al}{length}
 		if $tree->{fetchlength} < $cfg->{Alerts}{$al}{length};
@@ -277,23 +325,8 @@ sub init_target_tree ($$$$$$$$) {
 	    };
 	    init_target_tree $cfg, $probes, $tree->{probe}, $tree->{$prop}, "$name/$prop", $tree->{PROBE_CONF},$tree->{alerts},$tree->{alertee};
 	}
-        # if there is a filter do neither schedule these nor make rrds
-        if ($opt{filter} && scalar @{$opt{filter}}){
-            my $ok = 0;
-            for (@{$opt{filter}}){
-              /^\!(.+)$/ && do {
-                "$name" =~ /^$1/ && do{ $ok = 0};
-                next;
-              };
-              /^(.+)$/ && do {
-                "$name" =~ /^$1/ && do {$ok = 1};
-                next;
-              }; 
-            }  
-            next unless $ok; 
-        };
-
-	if ($prop eq 'host') {
+	if ($prop eq 'host' and check_filter($name)) {           
+	    # print "init $name\n";
 	    if (not -f $name.".rrd"){
 		RRDs::create
 			($name.".rrd", "--step",$cfg->{Database}{step},
@@ -551,9 +584,6 @@ sub findmax ($$) {
         } else {
                 $maxmedian{$x} = $max * 1.5;
         }
-        # make it a nice clean number
-        $maxmedian{$x} =~ s/^([0-9.]*)([1-9]).*/$1$2/;
-        $maxmedian{$x} += ( ${1}."1" + 0.0 ) ;
 
         $maxmedian{$x} = $cfg->{Presentation}{detail}{max_rtt} 
                 if $cfg->{Presentation}{detail}{max_rtt} and
@@ -835,26 +865,12 @@ sub update_rrds($$$$$) {
     $probe = $tree->{probe} if defined $tree->{probe};
     foreach my $prop (keys %{$tree}) {
 
-       if ($opt{filter} && scalar @{$opt{filter}}){
-            my $ok = 0;
-            for (@{$opt{filter}}){
-              /^\!(.+)$/ && do {
-                "$name/$prop" =~ /^$1/ && do{ $ok = 0};
-                next;
-              };
-              /^(.+)$/ && do {
-                "$name/$prop" =~ /^$1/ && do {$ok = 1};
-                next;
-              };
-            }
-            next unless $ok;
-        };
-
     	next if $prop eq "PROBE_CONF";
         if (ref $tree->{$prop} eq 'HASH'){
             update_rrds $cfg, $probes, $probe, $tree->{$prop}, $name."/$prop";
         } 
-        if ($prop eq 'host') {
+        if ($prop eq 'host' and check_filter($name)) {
+            #print "update $name\n";
 	    my $updatestring = $probes->{$probe}->rrdupdate_string($tree);
 	    if ( $tree->{rawlog} ){
 		my $file =  POSIX::strftime $tree->{rawlog},localtime(time);
@@ -909,13 +925,26 @@ sub update_rrds($$$$$) {
                         my $loss = "loss: ".join ", ",map {defined $_ ? (/^\d/ ? sprintf "%.0f%%", $_ / $cfg->{Database}{pings} * 100:$_):"U" } @{$x->{loss}};
                         my $rtt = "rtt: ".join ", ",map {defined $_ ? (/^\d/ ? sprintf "%.0fms", $_*1000 :$_):"U" } @{$x->{rtt}}; 
                         my $stamp = scalar localtime time;
-			if ($cfg->{Alerts}{to} =~ /^\|(.+)/) {
-			    system $1,$_,$line,$loss,$rtt,$tree->{host};	    			    
-			} else {
-			    my $to = $cfg->{Alerts}{to};
-     			    $to .= ",".$tree->{alertee} if $tree->{alertee};
+			my @to;
+			foreach my $addr (map {$_ ? (split /\s*,\s*/,$_) : ()} $cfg->{Alerts}{to},$tree->{alertee},$cfg->{Alerts}{$_}{to}){
+			     next unless $addr;
+			     if ( $addr =~ /^\|(.+)/) {
+  			         system $1,$_,$line,$loss,$rtt,$tree->{host};				     
+			     } elsif ( $addr =~ /^snpp:(.+)/ ) {
+				 sendsnpp $1, <<SNPPALERT;
+$cfg->{Alerts}{$_}{comment}
+$_ on $line
+$loss
+$rtt
+SNPPALERT
+			     } else {
+			    	 push @to, $addr;
+			     }
+			};
+			if (@to){
+			    my $to = join ",",@to;
 			    sendmail $cfg->{Alerts}{from},$to, <<ALERT;
-To: $cfg->{Alerts}{to}
+To: $to
 From: $cfg->{Alerts}{from}
 Subject: [SmokeAlert] $_ on $line
 
@@ -927,8 +956,8 @@ Pattern
 -------
 $cfg->{Alerts}{$_}{pattern}
 
-Data
-----
+Data (old --> now)
+------------------
 $loss
 $rtt
 
@@ -1038,11 +1067,11 @@ DOC
                         return undef;
         	  }, 
            },
-	   alertee => { _re => '(\S+@\S+)',
+	   alertee => { _re => '(\|.+|.+@\S+|snpp:)',
 			_re_error => 'the alertee must be an email address here',
 			_doc => <<DOC },
 If you want to have alerts for this target and all targets below it go to a particular address
-on top of the address already specified in the alert, you can add it here.
+on top of the address already specified in the alert, you can add it here. This can be a comma separated list of items.
 DOC
 
     };
@@ -1096,7 +1125,7 @@ General configuration values valid for the whole SmokePing setup.
 DOC
 	 _vars =>
 	 [ qw(owner imgcache imgurl datadir pagedir piddir sendmail offset
-              smokemail cgiurl mailhost contact 
+              smokemail cgiurl mailhost contact netsnpp
 	      syslogfacility syslogpriority) ],
 	 _mandatory =>
 	 [ qw(owner imgcache imgurl datadir piddir
@@ -1138,6 +1167,14 @@ and use perl's Net::SMTP module to send mail to DYNAMIC host owners (see below).
 DOC
           _sub => sub { require Net::SMTP ||return "ERROR: loading Net::SMTP"; return undef; }
 	 },
+	 snpphost  => 
+	 {
+	  _doc => <<DOC,
+If you have a SNPP (Simple Network Pager Protocol) server at hand, you can have alerts
+sent there too. Use the syntax B<snpp:someaddress> to use a snpp address in any place where you can use a mail address otherwhise.
+DOC
+          _sub => sub { require Net::SNPP ||return "ERROR: loading Net::SNPP"; return undef; }
+	 },
 
 	 contact  => 
 	 { _re => '\S+@\S+',
@@ -1668,11 +1705,15 @@ DOC
 	     _sections => [ '/[^\s,]+/' ],
 	     _vars => [ qw(to from) ],
 	     _mandatory => [ qw(to from)],
-	     to => { doc => 'Either an email address to send alerts to, or the name of a
-program to execute when an alert matches. To call a program, the first character of the B<to>
-value must be a pipe symbol "|". The program will the be called whenever an alert matches,
-using the following 5 arguments: B<name-of-alert>, B<target>, B<loss-pattern>, B<rtt-pattern>, B<hostname>.',
-			_re => '(\|.+|.+@\S+)',
+	     to => { doc => <<DOC,
+Either an email address to send alerts to, or the name of a program to
+execute when an alert matches. To call a program, the first character of the
+B<to> value must be a pipe symbol "|". The program will the be called
+whenever an alert matches, using the following 5 arguments:
+B<name-of-alert>, B<target>, B<loss-pattern>, B<rtt-pattern>, B<hostname>.
+You can also provide a comma separated list of addresses and programs.
+DOC
+			_re => '(\|.+|.+@\S+|snpp:)',
 			_re_error => 'put an email address or the name of a program here',
 		      },
 	     from => { doc => 'whom should alerts appear to be coming from ?',
@@ -1680,17 +1721,22 @@ using the following 5 arguments: B<name-of-alert>, B<target>, B<loss-pattern>, B
 		       _re_error => 'put an email address here',
 		      },
 	     '/[^\s,]+/' => {
-		  _vars => [ qw(type pattern comment) ],
+		  _vars => [ qw(type pattern comment to) ],
 		  _mandatory => [ qw(type pattern comment) ],
+	          to => { doc => 'Similar to the "to" parameter on the top-level except that  it will only be used IN ADDITION to the value of the toplevel parameter. Same rules apply.',
+			_re => '(\|.+|.+@\S+|snpp:)',
+			_re_error => 'put an email address or the name of a program here',
+		          },
+		  
 		  type => {
-		     _doc => 'Currently the pattern types B<rtt> and B<loss> are known',
-		     _re => '(rtt|loss)',
+		     _doc => 'Currently the pattern types B<rtt> and B<loss> and B<matcher> are known',
+		     _re => '(rtt|loss|matcher)',
                      _re_error => 'Use loss or rtt'
 			  },
    	 	  pattern => {
  		     _doc => "a comma separated list of comparison operators and numbers. rtt patterns are in milli seconds, loss patterns are in percent",
-		     _re => '([^,]+)(,[^,]+)*',
- 		     _re_error => 'Could not parse pattern',
+		     _re => '(?:([^,]+)(,[^,]+)*|\S+\(.+\s)',
+ 		     _re_error => 'Could not parse pattern or matcher',
 		             },
 		  },
         },
@@ -1748,7 +1794,7 @@ sub kill_smoke ($) {
         if ( -f $pidfile && open PIDFILE, "<$pidfile" ) {
             <PIDFILE> =~ /(\d+)/;
             my $pid = $1;
-            kill 1, $pid if kill 0, $pid;
+            kill 2, $pid if kill 0, $pid;
             sleep 3; # let it die
             die "ERROR: Can not stop running instance of SmokePing ($pid)\n"
                 if kill 0, $pid;    
@@ -1862,7 +1908,7 @@ sub daemonize_me ($) {
 # The Main Program 
 ###########################################################################
 
-my $RCS_VERSION = '$Id: Smokeping.pm,v 1.4 2002/03/09 12:01:08 oetiker Exp $';
+my $RCS_VERSION = '$Id: Smokeping.pm,v 1.5 2004/10/21 21:10:51 oetiker Exp $';
 
 sub load_cfg ($) { 
     my $cfgfile = shift;
@@ -1889,9 +1935,10 @@ sub makepod ($){
     my $cfg = shift;
     my $e='=';
     print <<POD;
+
 ${e}head1 NAME
 
-smokeping.config - Reference for the SmokePing Config File
+smokeping_config - Reference for the SmokePing Config File
 
 ${e}head1 OVERVIEW
 
@@ -2055,7 +2102,7 @@ sub main ($) {
     my $cfgfile = shift;
     $opt{filter}=[];
     GetOptions(\%opt, 'version', 'email', ,'man','help','logfile=s','static-pages:s',
-                      'makepod','debug','restart', 'filter=s', 'nodaemon|nodemon') or pod2usage(2);
+		      'nosleep', 'makepod','debug','restart', 'filter=s', 'nodaemon|nodemon') or pod2usage(2);
     if($opt{version})  { print "$RCS_VERSION\n"; exit(0) };
     if($opt{man})      {  pod2usage(-verbose => 2); exit 0 };
     if($opt{help})     {  pod2usage(-verbose => 1); exit 0 };
@@ -2096,7 +2143,7 @@ sub main ($) {
         do_log "WARNING: smokeping took $runtime seconds to complete 1 round of polling. ".
              "It should complete polling in $cfg->{Database}{step} seconds. ".
              "You may have unresponsive devices in your setup.\n" if $runtime > $cfg->{Database}{step};
-	sleep $cfg->{Database}{step} - (time+$offset) % $cfg->{Database}{step};
+	sleep $cfg->{Database}{step} - (time+$offset) % $cfg->{Database}{step} unless $opt{nosleep};
     }
 }
 
